import codecs
import requests


def get_translations(word, language_from, language_to):
    results = []

    params = dict({'action': 'query', 'prop': 'iwlinks', 'titles': word, 'iwprefix': language_to, 'format': 'json'})

    r = requests.get('https://' + language_from + '.wiktionary.org/w/api.php', params=params)
    j = r.json()
    for page in j['query']['pages'].values():
        if 'iwlinks' in page:
            for link in page['iwlinks']:
                results.append(link['*'])

    return results


def get_ergative_verbs():
    title = 'Category:Ergatief_werkwoord_in_het_Nederlands'
    with codecs.open('../data/nl_aux_be.txt', 'w', 'utf-8') as f:
        continue_url = ''

        params = dict({'action': 'query', 'list': 'categorymembers', 'cmtitle': title, 'format': 'json'})
        while True:
            params['cmcontinue'] = continue_url

            r = requests.get('https://nl.wiktionary.org/w/api.php', params=params)
            j = r.json()
            for member in j['query']['categorymembers']:
                f.write(member['title'] + '\n')

            # Continue to next URL
            if 'continue' not in j:
                break
            continue_url = j['continue']['cmcontinue']


if __name__ == "__main__":
    get_ergative_verbs()



import time

import click

from perfectextractor.corpora.bnc.extractor import BNCExtractor, BNCPerfectExtractor
from perfectextractor.corpora.dpc.extractor import DPCExtractor, DPCPerfectExtractor
from perfectextractor.corpora.europarl.extractor import EuroparlExtractor, EuroparlPoSExtractor, EuroparlPerfectExtractor, \
    EuroparlRecentPastExtractor, EuroparlSinceDurationExtractor
from perfectextractor.apps.extractor.utils import TXT, XML
from perfectextractor.apps.extractor.perfectextractor import PRESENT, PAST

# Corpora
BNC = 'bnc'
DPC = 'dpc'
EUROPARL = 'europarl'

# Extractor types
BASE = 'base'
POS = 'pos'
PERFECT = 'perfect'
RECENT_PAST = 'recent_past'
SINCE_DURATION = 'since_duration'


def process_data_folders(extractor, path):
    for directory in extractor.list_directories(path):
        t0 = time.time()
        click.echo('Now processing {} for {}'.format(directory, extractor.l_from))
        extractor.process_folder(directory)
        click.echo('Processing finished, took {:.3} seconds'.format(time.time() - t0))


@click.command()
@click.argument('folder')
@click.argument('language_from')
@click.argument('languages_to', nargs=-1)  # nargs=-1 eats up all remaining arguments
@click.option('--corpus', default=EUROPARL, type=click.Choice([EUROPARL, DPC, BNC]),
              help='Which type of corpus to use')
@click.option('--extractor', default=BASE, type=click.Choice([BASE, POS, PERFECT, RECENT_PAST, SINCE_DURATION]),
              help='Which kind of extractor to use')
@click.option('--file_names', '-f', multiple=True,
              help='Limits the file names searched into')
@click.option('--sentence_ids', '-s', multiple=True,
              help='Limits the sentence IDs searched into')
@click.option('--lemmata', '-l', multiple=True,
              help='Limits the lemmata searched for')
@click.option('--regex', '-r', multiple=True,
              help='Use regular expression to match words')
@click.option('--pos', '-p', multiple=True,
              help='Limits the POS-tags searched for')
@click.option('--tokens', '-t', multiple=True, type=click.Tuple([str, str]),
              help='Limits the tokens searched for. Format: -t [start_token] [end_token]')
@click.option('--metadata', '-m', multiple=True, type=click.Tuple([str, str]),
              help='Adds additional metadata. Format: -m [tag] [level]')
@click.option('--outfile', '-o',
              help='Output file')
@click.option('--position', default=0,
              help='The position of the searched item')
@click.option('--search_in_to', is_flag=True,
              help='Also search for perfects in the to language(s)?')
@click.option('--tense', default=PRESENT, type=click.Choice([PRESENT, PAST]),
              help='The tense of perfect (present, past, future)')
@click.option('--output', default=TXT, type=click.Choice([TXT, XML]),
              help='Output in text or XML format')
@click.option('--one_per_sentence', is_flag=True,
              help='Output all sentences, and only one classification per sentence')
@click.option('--sort_by_certainty', is_flag=True,
              help='Sort by certainty?')
@click.option('--file_limit', default=0,
              help='Limit number of files searched')
@click.option('--min_file_size', default=0,
              help='Limits the minimal size of the files searched')
@click.option('--max_file_size', default=0,
              help='Limits the maximal size of the files searched')
def extract(folder, language_from, languages_to, corpus='europarl', extractor='base',
            pos=None, search_in_to=False, tense=PRESENT,
            output=TXT, file_names=None, sentence_ids=None,
            lemmata=None, regex=None, position=None, tokens=None, metadata=None,
            outfile=None, one_per_sentence=False, sort_by_certainty=False, file_limit=0,
            min_file_size=0, max_file_size=0):
    # Set the default arguments
    kwargs = dict(output=output, file_names=file_names, sentence_ids=sentence_ids,
                  lemmata=lemmata, regex=regex, position=position, tokens=tokens, metadata=metadata,
                  outfile=outfile, one_per_sentence=one_per_sentence,
                  sort_by_certainty=sort_by_certainty, file_limit=file_limit,
                  min_file_size=min_file_size, max_file_size=max_file_size)

    # Determine the extractor to be used
    # TODO: add more varieties
    resulting_extractor = None
    if corpus == EUROPARL:
        if extractor == POS:
            resulting_extractor = EuroparlPoSExtractor
        elif extractor == PERFECT:
            resulting_extractor = EuroparlPerfectExtractor
        elif extractor == RECENT_PAST:
            resulting_extractor = EuroparlRecentPastExtractor
        elif extractor == SINCE_DURATION:
            resulting_extractor = EuroparlSinceDurationExtractor
        else:
            resulting_extractor = EuroparlExtractor
    elif corpus == DPC:
        if extractor == POS:
            raise click.ClickException('Corpus or extractor type not implemented!')
        elif extractor == PERFECT:
            resulting_extractor = DPCPerfectExtractor
        elif extractor == RECENT_PAST:
            raise click.ClickException('Corpus or extractor type not implemented!')
        elif extractor == SINCE_DURATION:
            raise click.ClickException('Corpus or extractor type not implemented!')
        else:
            resulting_extractor = DPCExtractor
    elif corpus == BNC:
        if extractor == POS:
            raise click.ClickException('Corpus or extractor type not implemented!')
        elif extractor == PERFECT:
            resulting_extractor = BNCPerfectExtractor
        elif extractor == RECENT_PAST:
            raise click.ClickException('Corpus or extractor type not implemented!')
        elif extractor == SINCE_DURATION:
            raise click.ClickException('Corpus or extractor type not implemented!')
        else:
            resulting_extractor = BNCExtractor

    if extractor == PERFECT:
        kwargs['search_in_to'] = search_in_to
        kwargs['tense'] = tense

    if extractor == POS:
        kwargs['pos'] = pos

    if not resulting_extractor:
        raise click.ClickException('Unknown value for either corpus or extractor type')

    # Start the extraction!
    process_data_folders(resulting_extractor(language_from, languages_to, **kwargs), folder)


if __name__ == "__main__":
    extract()


# TODO:
# switch between perfect, recentpast, pos, word
# dependent on this choice, add other questions
# allow output to .xlsx



import unittest

from perfectextractor.apps.extractor.utils import get_adjacent_line_number
from perfectextractor.corpora.dpc.utils import is_nl


class TestUtils(unittest.TestCase):
    def test_get_adjacent_line_number(self):
        next_line = get_adjacent_line_number('p1.s1', 2)
        self.assertEqual(next_line, 'p1.s3')
        next_line = get_adjacent_line_number('p2.s6', 4)
        self.assertEqual(next_line, 'p2.s10')

    def test_is_nl(self):
        self.assertEqual(is_nl('nl'), 1)
        self.assertEqual(is_nl('en'), 0)



import unittest

from perfectextractor.apps.extractor.models import PresentPerfect

XML_ID = 'test_id'


class TestPresentPerfect(unittest.TestCase):
    def setUp(self):
        self.pp = PresentPerfect('has', 'have', XML_ID)
        self.pp.add_word('always', 'always', False, XML_ID)
        self.pp.add_word('loved', 'love', True, XML_ID)

    def test_extractions(self):
        self.assertEqual(self.pp.perfect_lemma(), 'love')
        self.assertEqual(self.pp.verbs(), ['has', 'loved'])
        self.assertEqual(self.pp.verbs_to_string(), 'has loved')
        self.assertEqual(self.pp.words_between(), 1)

    def test_extend(self):
        ppp = PresentPerfect('has', 'have', XML_ID)
        ppp.add_word('been', 'be', True, XML_ID)
        pp_extend = PresentPerfect('been', 'be', XML_ID)
        pp_extend.add_word('created', 'create', True, XML_ID)
        ppp.extend(pp_extend)

        self.assertEqual(ppp.perfect_lemma(), 'create')
        self.assertEqual(ppp.verbs(), ['has', 'been', 'created'])
        self.assertEqual(ppp.verbs_to_string(), 'has been created')
        self.assertEqual(ppp.words_between(), 0)



from abc import ABCMeta
import string

from .base import BaseExtractor
from .models import MultiWordExpression

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser


class RecentPastExtractor(BaseExtractor):
    __metaclass__ = ABCMeta

    def __init__(self, language_from, languages_to=None, **kwargs):
        """
        Initializes the extractor for the given source and target language(s).
        Reads in the config for the source language.
        :param language_from: the source language
        :param languages_to: the target language(s)
        """
        super(RecentPastExtractor, self).__init__(language_from, languages_to, **kwargs)

    def check_recent_past(self, w, language):
        """
        Checks if the element w is the start of a recent past construction
        :param w: the starting element
        :param language: the language
        :return: if found, the recent past construction as a MultiWordExpression, otherwise None
        """
        is_recent_past = False

        # Retrieve the configuration variables
        lemma_attr = self.config.get('all', 'lemma_attr')
        pos_attr = self.config.get(language, 'pos')
        rp_pre_pos = self.config.get(language, 'rp_pre_pos').split(',')
        rp_pre_lem = self.config.get(language, 'rp_pre_lem')
        rp_inf_pos = self.config.get(language, 'rp_inf_pos')
        check_ppp = self.config.get(language, 'ppp')
        ppp_lemma = self.config.get(language, 'ppp_lemma')
        perfect_tags = self.config.get(language, 'perfect_tags').split(',')
        stop_tags = tuple(self.config.get(language, 'stop_tags').split(','))

        sentence = self.get_sentence(w)

        # Start a new MWE at the first word
        mwe = MultiWordExpression(sentence)
        mwe.add_word(w.text, w.get(lemma_attr), True, w.get('id'))

        # Check the siblings for the preposition of the recent past construction
        for pre in self.get_siblings(w, sentence.get('id'), False):
            pre_pos = pre.get(pos_attr)
            pre_lem = pre.get(lemma_attr)
            if pre_pos in rp_pre_pos and pre_lem == rp_pre_lem:
                mwe.add_word(pre.text, pre_lem, True, pre.get('id'))

                # Now look at the siblings to find an infinitive
                for inf in self.get_siblings(pre, sentence.get('id'), False):
                    inf_pos = inf.get(pos_attr)
                    inf_lem = inf.get(lemma_attr)
                    if inf_pos == rp_inf_pos:
                        is_recent_past = True
                        mwe.add_word(inf.text, inf_lem, True, inf.get('id'))

                        # If the language has passive recent pasts, check if this is followed by a perfect
                        if check_ppp and inf_lem == ppp_lemma:
                            s_next = inf.getnext()
                            if s_next.get(pos_attr) in perfect_tags:
                                mwe.add_word(s_next.text, s_next.get(lemma_attr), True, s_next.get('id'))

                        # Break out of the loop: we found our recent past construction
                        break
                    # Stop looking at punctuation or stop tags
                    elif (inf.text and inf.text in string.punctuation) or (inf_pos and inf_pos.startswith(stop_tags)):
                        break
                    # Otherwise: add the word to the MWE
                    else:
                        mwe.add_word(inf.text, inf_lem, False, inf.get('id'))

                # If we found our recent past construction: break out of the loop
                if is_recent_past:
                    break
            # Stop looking at punctuation or stop tags
            elif (pre.text and pre.text in string.punctuation) or (pre_pos and pre_pos.startswith(stop_tags)):
                break
            # Otherwise: add the word to the MWE
            else:
                mwe.add_word(pre.text, pre_lem, False, pre.get('id'))

        return mwe if is_recent_past else None



from abc import ABCMeta, abstractmethod

from perfectextractor.apps.extractor.utils import UnicodeWriter


class BaseCounter(object):
    __metaclass__ = ABCMeta

    def __init__(self, language_from):
        """
        Initializes the counter for the given source and target language(s).
        :param language_from: the source language
        """
        self.l_from = language_from

    def process_folder(self, dir_name):
        """
        Creates a result file and processes each file in a folder.
        """
        result_file = '-counts-'.join([dir_name, self.l_from]) + '.csv'
        with open(result_file, 'w') as f:
            f.write(u'\uFEFF'.encode('utf-8'))  # the UTF-8 BOM to hint Excel we are using that...
            csv_writer = UnicodeWriter(f, delimiter=';')

            header = ['document', 'lemma', 'count']
            csv_writer.writerow(header)

            for filename in self.list_filenames(dir_name):
                results = self.process_file(filename)
                csv_writer.writerows(results)

    @abstractmethod
    def get_config(self):
        """
        Returns the location of the configuration file.
        """
        raise NotImplementedError

    @abstractmethod
    def process_file(self, filename):
        """
        Process a single file
        """
        raise NotImplementedError

    @abstractmethod
    def list_filenames(self, dir_name):
        """
        List all to be processed files in the given directory.
        """
        raise NotImplementedError



# -*- encoding: utf-8 -*-

import os

from lxml import etree

from perfectextractor.apps.extractor.base import BaseExtractor
from perfectextractor.apps.extractor.perfectextractor import PerfectExtractor
from perfectextractor.apps.extractor.utils import XML

from .base import BaseBNC


class BNCExtractor(BaseBNC, BaseExtractor):
    def process_file(self, filename):
        results = []

        # Retrieve the genre
        tree = etree.parse(filename)
        genre = self.get_genre(tree)

        # Parse the current tree (create a iterator over 's' elements)
        s_trees = etree.iterparse(filename, tag='s')

        # Find potential present perfects
        for _, s in s_trees:
            if self.sentence_ids and s.get('n') not in self.sentence_ids:
                continue

            result = list()
            result.append(os.path.basename(filename))
            result.append(s.get('n'))
            result.append(genre)
            result.append('')
            result.append('')
            if self.output == XML:
                result.append('<root>' + etree.tostring(s) + '</root>')
            else:
                result.append(self.get_sentence_words(s))

            results.append(result)
        return results

    def get_sentence(self, element):
        return element.xpath('ancestor::s')[0]

    def get_siblings(self, element, sentence_id, check_preceding):
        return element.itersiblings(tag='w', preceding=check_preceding)

    def get_sentence_words(self, sentence):
        """
        Returns all words in the sentence.
        Note that in the BNC, these can be either in w or c tags (the latter are for punctuation)
        :param sentence: the s element
        :return: all w and c texts, joined with a space.
        """
        s = []
        for w in sentence.xpath('.//w | .//c'):
            s.append(w.text.strip() if w.text else ' ')
        return ' '.join(s)

    def sort_by_alignment_certainty(self, file_names):
        raise NotImplementedError

    def filter_by_file_size(self, file_names):
        raise NotImplementedError

    def get_translated_lines(self, alignment_trees, language_from, language_to, segment_number):
        raise NotImplementedError


class BNCPerfectExtractor(BNCExtractor, PerfectExtractor):
    def get_line_and_pp(self, tree, language_to, segment_number):
        raise NotImplementedError

    def generate_header(self):
        header = [
            'document',
            'genre',
            'is-perfect',
            'tense',
            'words',
            'lemma',
            'is-question',
            'text']
        return header

    def process_file(self, filename):
        """
        Processes a single file.
        """
        results = []

        # Retrieve the genre
        tree = etree.parse(filename)
        genre = self.get_genre(tree)

        # if not genre.startswith('S'):  # Only spoken genre for the moment
        #    return results

        # Parse the current tree (create a iterator over 's' elements)
        s_trees = etree.iterparse(filename, tag='s')

        # Find potential present perfects
        for _, s in s_trees:
            sentence = self.get_sentence_words(s)
            is_question = self.is_question(sentence)

            for e in s.xpath(self.config.get(self.l_from, 'xpath')):
                pp = self.check_present_perfect(e, self.l_from)

                # If this is really a present perfect, add it to the result
                if pp:
                    result = list()
                    result.append(os.path.basename(filename))
                    result.append(genre)
                    result.append('1')
                    result.append(pp.perfect_type())
                    result.append(pp.verbs_to_string())
                    result.append(pp.perfect_lemma())
                    result.append('1' if is_question else '0')
                    result.append(sentence)
                    results.append(result)

                    # If we want (only) one classification per sentence, break the for loop here.
                    if self.one_per_sentence:
                        break
            else:
                # If we want one classification per sentence, add the sentence with a classification here.
                if self.one_per_sentence:
                    tense, tenses = self.get_tenses(s)

                    result = list()
                    result.append(os.path.basename(filename))
                    result.append(genre)
                    result.append('0')
                    result.append(tense)
                    result.append(','.join(tenses))
                    result.append('')
                    result.append('1' if is_question else '0')
                    result.append(sentence)
                    results.append(result)

        return results

    def is_question(self, sentence):
        """
        Naively checks if a sentence is a question.
        :param sentence: a string of words separated by spaces
        :return: whether this sentence contains a question mark
        """
        return '?' in sentence

    def get_pos(self, language, element):
        """
        POS-tags in the BNC can potentially consist of two part-of-speech tags,
        but the first tag is the most likely choice.
        See http://www.natcorp.ox.ac.uk/docs/URG/posguide.html#ambiguity
        :param language: the current language
        :param element: the current w element
        :return: the (most likely) part-of-speech tag
        """
        pos = element.get(self.config.get(language, 'pos'))
        return pos.split('-')[0]



def get_original_language(element):
    """
    Returns the original language for a document.
    """
    speaker_language = element.getparent().getparent().getparent().get('LANGUAGE')
    return speaker_language or '?'


def get_sentence_from_element(element):
    s = []
    for w in element.xpath('ancestor::s')[0].xpath('.//w'):
        s.append(w.text)
    return ' '.join(s)



import glob
import os

from lxml import etree

from perfectextractor.apps.extractor.base import BaseExtractor
from perfectextractor.apps.extractor.perfectextractor import PerfectExtractor
from .utils import is_nl, NL

DPC_CONFIG = os.path.join(os.path.dirname(__file__), 'dpc.cfg')
TEI = {'ns': 'http://www.tei-c.org/ns/1.0'}


class DPCExtractor(BaseExtractor):
    def get_config(self):
        return DPC_CONFIG

    def process_file(self, filename):
        raise NotImplementedError

    def list_filenames(self, dir_name):
        return sorted(glob.glob(os.path.join(dir_name, '*[0-9]-' + self.l_from + '-tei.xml')))

    def get_translated_lines(self, alignment_trees, language_from, language_to, segment_number):
        """
        Returns the translated segment numbers (could be multiple) for a segment number in the original text.

        The translation document file format either ends with nl-en-tei.xml or nl-fr-tei.xml.

        Alignment lines look like this:
            <link type="A: 1-1" targets="p1.s1; p1.s1"/>
            <link type="A: 1-2" targets="p1.s9; p1.s9 p1.s10"/>
            <link type="B: 2-1" targets="p1.s24 p1.s25; p1.s25"/>

        To get from NL to EN/FR, we have to find the segment number in the targets attribute BEFORE the semicolon.
        For the reverse pattern, we have to find the segment number in the targets attribute AFTER the semicolon.

        To get from EN to FR or from FR to EN, we have to use NL as an in between language.

        This function supports 1-to-2 alignments, as it will return the translated lines as a list.

        TODO: deal with 2-to-2 and 2-to-1 alignments as well here.
        """
        result = []
        alignment_type = ''

        if NL in [language_from, language_to]:
            not_nl = language_to if language_to != NL else language_from
            for link in alignment_trees[not_nl].xpath('//ns:link', namespaces=TEI):  # TODO: simplify this
                alignment_type = link.get('type').split(': ')[1]
                if is_nl(language_to):
                    alignment_type = alignment_type[::-1]  # reverse the alignment type
                alignment_type = alignment_type.replace('-', '=>')
                targets = link.get('targets').split('; ')
                if segment_number in targets[1 - is_nl(language_from)].split(' '):
                    result = targets[is_nl(language_from)].split(' ')
                    break
        else:
            lookup, t = self.get_translated_lines(alignment_trees, language_from, NL, segment_number)
            alignment_type = t.split('=>')[0]
            for lookup_number in lookup:
                lines, _ = self.get_translated_lines(alignment_trees, NL, language_to, lookup_number)
                result.extend(lines)

            if result:
                alignment_type += '=>' + str(len(set(result)))
            else:
                alignment_type = ''

        if not result:
            alignment_type = ''

        return set(result), alignment_type

    def get_sentence(self, element):
        return element.xpath('ancestor::ns:s', namespaces=TEI)[0]

    def get_siblings(self, element, sentence_id, check_preceding):
        return element.itersiblings(preceding=check_preceding)

    def sort_by_alignment_certainty(self, file_names):
        raise NotImplementedError

    def filter_by_file_size(self, file_names):
        raise NotImplementedError


class DPCPerfectExtractor(PerfectExtractor, DPCExtractor):
    def get_line_and_pp(self, tree, language_to, segment_number):
        """
        Returns the full line for a segment number, as well as the PresentPerfect found (or None if none found).
        TODO: handle more than one here? => bug
        """
        sentence = '-'
        pp = None

        line = tree.xpath('//ns:s[@n="' + segment_number + '"]', namespaces=TEI)
        if line:
            s = line[0]
            sentence = s.getprevious().text

            if self.search_in_to:
                for e in s.xpath(self.config.get(language_to, 'xpath'), namespaces=TEI):
                    pp = self.check_present_perfect(e, language_to)
                    if pp:
                        sentence = pp.mark_sentence()
                        break

        return sentence, pp

    def get_original_language(self, document):
        """
        Returns the original language for a document.
        """
        metadata_tree = etree.parse(document + self.l_from + '-mtd.xml')
        return metadata_tree.getroot().find('metaTrans').find('Original').get('lang')

    def process_file(self, filename):
        """
        Processes a single file.
        """
        document = filename.split(self.l_from + '-tei.xml')[0]
        results = []

        # Parse the current tree
        tree = etree.parse(filename)

        # Parse the trees with the translations
        translation_trees = dict()
        alignment_trees = dict()
        for language_to in self.l_to:
            translation_file = document + language_to + '-tei.xml'
            if os.path.exists(translation_file):
                translation_trees[language_to] = etree.parse(translation_file)

            not_nl = language_to if language_to != NL else self.l_from
            alignment_file = document + NL + '-' + not_nl + '-tei.xml'
            if os.path.isfile(alignment_file):
                alignment_trees[not_nl] = etree.parse(alignment_file)

        # Find potential present perfects
        for e in tree.xpath('//' + self.config.get(self.l_from, 'xpath'), namespaces=TEI):
            pp = self.check_present_perfect(e, self.l_from)

            # If this is really a present perfect, add it to the result
            if pp:
                result = list()
                result.append(document[:-1])
                result.append(self.get_original_language(document))
                result.append(pp.verbs_to_string())

                # Write the complete segment with mark-up
                result.append(pp.mark_sentence())

                # Find the translated lines
                segment_number = e.getparent().getparent().get('n')[4:]
                for language_to in self.l_to:
                    if language_to in translation_trees:
                        translated_lines, alignment_type = self.get_translated_lines(alignment_trees, self.l_from,
                                                                                     language_to, segment_number)
                        translated_present_perfects, translated_sentences = \
                            self.find_translated_present_perfects(translation_trees[language_to], language_to, translated_lines)
                        result.append('\n'.join([tpp.verbs_to_string() if tpp else '' for tpp in translated_present_perfects]))
                        result.append('\n'.join(self.check_translated_pps(pp, translated_present_perfects, language_to)))
                        result.append(alignment_type)
                        result.append('\n'.join(translated_sentences))
                    else:
                        # If no translation is available, add empty columns
                        result.extend([''] * 4)

                results.append(result)

        return results



from collections import Counter

from lxml import etree

from perfectextractor.apps.counter.base import BaseCounter
from .base import BaseEuroparl


class EuroparlCounter(BaseEuroparl, BaseCounter):
    def process_file(self, filename):
        """
        Processes a single file.
        """
        results = []

        # Parse the current tree
        tree = etree.parse(filename)

        c = Counter()
        for w in tree.xpath('.//w[starts-with(@tree, "V")]'):
            c[w.get('lem', '-')] += 1

        for k, v in c.most_common():
            results.append([filename, 'europarl', k, str(v)])

        return results



# -*- coding: utf-8 -*-

import unittest

from requests.exceptions import ConnectionError

from perfectextractor.apps.extractor.wiktionary import get_translations


class TestWiktionary(unittest.TestCase):
    def test_translations(self):
        try:
            self.assertIn('prove', get_translations('aantonen', 'nl', 'en'))
            self.assertIn('blijken', get_translations('prove', 'en', 'nl'))
            self.assertIn('arriver', get_translations('arriveren', 'nl', 'fr'))
        except ConnectionError as e:
            # No connection available, skipping tests
            pass



# -*- coding: utf-8 -*-

import unittest

from lxml import etree

from perfectextractor.apps.extractor.models import PresentPerfect
from perfectextractor.corpora.europarl.extractor import EuroparlPerfectExtractor


class TestPerfectExtractor(unittest.TestCase):
    def test_is_lexically_bound(self):
        extractor = EuroparlPerfectExtractor('en', ['de', 'es', 'fr', 'nl'], search_in_to=True)
        lemma_attr = extractor.config.get('all', 'lemma_attr')
        mock_pp = PresentPerfect('is', 'be', 'w1.1.1')

        # Default case: English (no lexical bounds)
        mock_aux_verb = {lemma_attr: 'have'}
        mock_participle = {lemma_attr: 'collided'}
        self.assertTrue(extractor.is_lexically_bound('en', mock_pp, mock_aux_verb, mock_participle))

        # Checking Dutch (ik ben gebotst vs. *ik ben gehad)
        mock_aux_verb = {lemma_attr: 'zijn'}
        mock_participle = {lemma_attr: 'botsen'}
        self.assertTrue(extractor.is_lexically_bound('nl', mock_pp, mock_aux_verb, mock_participle))

        mock_aux_verb = {lemma_attr: 'zijn'}
        mock_participle = {lemma_attr: 'hebben'}
        self.assertFalse(extractor.is_lexically_bound('nl', mock_pp, mock_aux_verb, mock_participle))

        # Checking French (*je suis regardé vs. je suis revenu)
        mock_aux_verb = {lemma_attr: u'être'}
        mock_participle = {lemma_attr: 'regarder'}
        self.assertFalse(extractor.is_lexically_bound('fr', mock_pp, mock_aux_verb, mock_participle))

        mock_aux_verb = {lemma_attr: u'être'}
        mock_participle = {lemma_attr: 'revenir'}
        self.assertTrue(extractor.is_lexically_bound('fr', mock_pp, mock_aux_verb, mock_participle))

        # Checking reflexive passé composés (je me suis couché)
        mock_pp = PresentPerfect('suis', u'être', 'w1.1.1')
        mock_aux_verb = {lemma_attr: u'être'}
        mock_participle = {lemma_attr: 'coucher'}
        self.assertFalse(extractor.is_lexically_bound('fr', mock_pp, mock_aux_verb, mock_participle))

        mock_sentence = etree.Element('s')
        me = etree.SubElement(mock_sentence, 'w')
        me.text = 'me'
        me.set(lemma_attr, 'me')
        me.set('pos', 'PRO:PER')
        je = etree.SubElement(mock_sentence, 'w')
        je.text = 'je'
        je.set(lemma_attr, 'je')
        je.set('pos', 'PRO:PER')

        self.assertTrue(extractor.is_lexically_bound('fr', mock_pp, mock_aux_verb, mock_participle, mock_sentence))
        self.assertEqual(len(mock_pp.words), 2)
        self.assertEqual(mock_pp.verbs_to_string(), 'me suis')
        self.assertTrue(mock_pp.is_reflexive)

        # Checking reflexive passé composés (puis nous sommes restés)
        mock_pp = PresentPerfect('sommes', u'être', 'w1.1.1')
        mock_aux_verb = {lemma_attr: u'être'}
        mock_participle = {lemma_attr: 'rester'}
        self.assertTrue(extractor.is_lexically_bound('fr', mock_pp, mock_aux_verb, mock_participle))

        mock_sentence = etree.Element('s')
        nous = etree.SubElement(mock_sentence, 'w')
        nous.text = 'nous'
        nous.set(lemma_attr, 'nous')
        nous.set('pos', 'PRO:PER')
        puis = etree.SubElement(mock_sentence, 'w')
        puis.text = 'puis'
        puis.set(lemma_attr, 'puis')
        puis.set('pos', 'ADV')

        # This should be lexically bound, but 'nous' should not be part of the passé composé
        self.assertTrue(extractor.is_lexically_bound('fr', mock_pp, mock_aux_verb, mock_participle, mock_sentence))
        self.assertEqual(mock_pp.verbs_to_string(), 'sommes')
        self.assertEqual(len(mock_pp.words), 1)
        self.assertFalse(mock_pp.is_reflexive)



# -*- coding: utf-8 -*-

import os
import unittest

from lxml import etree

from perfectextractor.corpora.europarl.extractor import EuroparlExtractor, EuroparlPerfectExtractor, EuroparlRecentPastExtractor, \
    EuroparlPoSExtractor, EuroparlSinceDurationExtractor, EuroparlFrenchArticleExtractor
from perfectextractor.apps.extractor.perfectextractor import PAST

EUROPARL_DATA = os.path.join(os.path.dirname(__file__), 'data/europarl')
DCEP_DATA = os.path.join(os.path.dirname(__file__), 'data/dcep')
SWITCHBOARD_DATA = os.path.join(os.path.dirname(__file__), 'data/switchboard')


class TestEuroparlPerfectExtractor(unittest.TestCase):
    def setUp(self):
        self.nl_filename = os.path.join(EUROPARL_DATA, 'nl/ep-00-12-15.xml')
        self.en_filename = os.path.join(EUROPARL_DATA, 'en/ep-00-12-15.xml')
        self.fr_filename = os.path.join(EUROPARL_DATA, 'fr/ep-00-12-15.xml')

        self.nl_extractor = EuroparlPerfectExtractor('nl', ['en'], search_in_to=True)
        self.nl_tree = etree.parse(self.nl_filename)
        self.nl_alignmenttrees, self.nl_translationtrees = self.nl_extractor.parse_alignment_trees(self.nl_filename)

        self.en_extractor = EuroparlPerfectExtractor('en', ['nl'], search_in_to=True)
        self.en_tree = etree.parse(self.en_filename)
        self.en_alignmenttrees, self.en_translationtrees = self.en_extractor.parse_alignment_trees(self.en_filename)

    def merge_results(self, generator):
        return sum(list(generator), [])

    def test_init(self):
        self.assertEqual(self.nl_extractor.config.get('nl', 'perfect_tags'), 'verbpapa')
        self.assertIn('dunken', self.nl_extractor.aux_be_list['nl'])

    def test_get_translated_lines(self):
        from_lines, to_lines, align = self.nl_extractor.get_translated_lines(self.nl_alignmenttrees, 'nl', 'en', '17')
        self.assertEqual(from_lines, ['17'])
        self.assertEqual(to_lines, ['11'])
        self.assertEqual(align, '1 => 1')

        from_lines, to_lines, align = self.nl_extractor.get_translated_lines(self.nl_alignmenttrees, 'nl', 'en', '18')
        self.assertEqual(from_lines, ['18', '19'])
        self.assertEqual(to_lines, ['12'])
        self.assertEqual(align, '2 => 1')

        from_lines, to_lines, align = self.nl_extractor.get_translated_lines(self.nl_alignmenttrees, 'nl', 'en', '57')
        self.assertEqual(from_lines, ['57'])
        self.assertEqual(to_lines, ['46', '47'])
        self.assertEqual(align, '1 => 2')

        from_lines, to_lines, align = self.nl_extractor.get_translated_lines(self.nl_alignmenttrees, 'nl', 'en', '9')
        self.assertEqual(from_lines, ['9'])
        self.assertEqual(to_lines, [])
        self.assertEqual(align, '')

        from_lines, to_lines, align = self.en_extractor.get_translated_lines(self.en_alignmenttrees, 'en', 'nl', '19')
        self.assertEqual(from_lines, ['19'])
        self.assertEqual(to_lines, ['27'])
        self.assertEqual(align, '1 => 1')

        from_lines, to_lines, align = self.en_extractor.get_translated_lines(self.en_alignmenttrees, 'en', 'nl', '8')
        self.assertEqual(from_lines, ['8'])
        self.assertEqual(to_lines, ['13', '14'])
        self.assertEqual(align, '1 => 2')

        from_lines, to_lines, align = self.en_extractor.get_translated_lines(self.en_alignmenttrees, 'en', 'nl', '234')
        self.assertEqual(from_lines, ['234', '235'])
        self.assertEqual(to_lines, ['290'])
        self.assertEqual(align, '2 => 1')

    def test_get_line_by_number(self):
        xml_sentence, _, pp = self.nl_extractor.get_line_and_pp(self.nl_tree, 'nl', '4')
        self.assertEqual(etree.fromstring(xml_sentence).get('id'), '4')
        self.assertEqual(pp.get_sentence_id(), '4')
        self.assertEqual(pp.verbs(), ['is', 'aangebroken'])
        self.assertEqual(pp.verb_ids(), 'w4.9 w4.19')
        self.assertEqual(pp.words_between(), 9)
        self.assertFalse(pp.is_passive)
        self.assertFalse(pp.is_continuous)

        xml_sentence, _, pp = self.nl_extractor.get_line_and_pp(self.nl_tree, 'nl', '15')
        self.assertEqual(etree.fromstring(xml_sentence).get('id'), '15')
        self.assertEqual(pp.verbs(), ['heeft', 'bemoeid'])
        self.assertEqual(pp.words_between(), 0)
        self.assertFalse(pp.is_passive)
        self.assertFalse(pp.is_continuous)

        xml_sentence, _, pp = self.nl_extractor.get_line_and_pp(self.nl_translationtrees['en'], 'en', '6')
        self.assertEqual(etree.fromstring(xml_sentence).get('id'), '6')
        self.assertEqual(pp.verbs(), ['has', 'said'])
        self.assertEqual(pp.words_between(), 1)
        self.assertFalse(pp.is_passive)
        self.assertFalse(pp.is_continuous)

        xml_sentence, _, pp = self.en_extractor.get_line_and_pp(self.en_tree, 'en', '89')
        self.assertEqual(etree.fromstring(xml_sentence).get('id'), '89')
        self.assertEqual(pp.verbs(), ['has', 'been', 'mentioned'])
        self.assertEqual(pp.words_between(), 1)
        self.assertEqual(pp.words_between_verbs(), [0, 1, 0])
        self.assertTrue(pp.is_passive)
        self.assertFalse(pp.is_continuous)

        xml_sentence, _, pp = self.en_extractor.get_line_and_pp(self.en_tree, 'en', '121')
        self.assertEqual(etree.fromstring(xml_sentence).get('id'), '121')
        self.assertEqual(pp.verbs(), ['has', 'been', 'carrying'])
        self.assertEqual(pp.words_between(), 0)
        self.assertEqual(pp.words_between_verbs(), [0, 0, 0])
        self.assertFalse(pp.is_passive)
        self.assertTrue(pp.is_continuous)

        xml_sentence, _, pp = self.en_extractor.get_line_and_pp(self.en_tree, 'en', '180')
        self.assertEqual(etree.fromstring(xml_sentence).get('id'), '180')
        self.assertEqual(pp.verbs(), ['has', 'brought'])
        self.assertEqual(pp.words_between(), 1)
        self.assertFalse(pp.is_passive)
        self.assertFalse(pp.is_continuous)

    def test_list_filenames(self):
        files = self.nl_extractor.list_filenames(os.path.join(EUROPARL_DATA, 'nl'))
        self.assertEqual([os.path.basename(f) for f in files], ['ep-00-12-15.xml'])

    def test_recent_past_extraction(self):
        fr_extractor = EuroparlRecentPastExtractor('fr', ['en', 'nl'])

        results = list(fr_extractor.process_file(self.fr_filename))
        self.assertEqual(len(results), 4)
        self.assertEqual(results[0][3], u'vient de dire')
        self.assertEqual(results[1][3], u'viens d\' aborder')
        self.assertEqual(results[2][3], u'viens d\' évoquer')
        self.assertEqual(results[3][3], u'vient d\' être dit')

    def test_position(self):
        when_extractor = EuroparlPoSExtractor('en', ['nl'], lemmata=['when'], position=1)
        results = self.merge_results(when_extractor.generate_results(os.path.join(EUROPARL_DATA, 'en')))
        self.assertEqual(len(results), 3)

    def test_average_alignment_certainty(self):
        extractor = EuroparlExtractor('en', ['nl', 'de'])

        file_names = extractor.sort_by_alignment_certainty(extractor.list_filenames(os.path.join(DCEP_DATA, 'en')))
        file_names = [os.path.basename(f) for f in file_names]
        self.assertEqual(file_names[0], '20764633__IM-PRESS__20081211-STO-44307__EN.xml')
        self.assertEqual(file_names[1], '16609396__IM-PRESS__20060905-STO-10339__EN.xml')
        self.assertEqual(file_names[2], '16451293__IM-PRESS__20060131-IPR-04891__EN.xml')

    def test_file_limit(self):
        extractor = EuroparlExtractor('en', ['nl', 'de'], file_limit=2)
        results = self.merge_results(extractor.generate_results(os.path.join(DCEP_DATA, 'en')))
        self.assertEqual(len(results), 55)

        extractor = EuroparlExtractor('en', ['nl', 'de'], file_limit=1)
        results = self.merge_results(extractor.generate_results(os.path.join(DCEP_DATA, 'en')))
        self.assertEqual(len(results), 37)

    def test_regex(self):
        # Primitive search for wh-questions
        regex_extractor = EuroparlPoSExtractor('en', ['nl'], regex=['^wh.*', '^how$'], position=1)
        results = self.merge_results(regex_extractor.generate_results(os.path.join(EUROPARL_DATA, 'en')))
        self.assertEqual(len(results), 12)
        self.assertEqual(results[0][3], u'How')
        self.assertEqual(results[1][3], u'How')
        self.assertEqual(results[2][3], u'What')
        self.assertEqual(results[3][3], u'When')

    def test_tokens(self):
        tokens_extractor = EuroparlPoSExtractor('en', ['nl'], tokens=[('w1.13', 'w1.15'), ('w2.5', 'w2.8')])
        results = self.merge_results(tokens_extractor.generate_results(os.path.join(EUROPARL_DATA, 'en')))
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0][3], u'a historic sitting')
        self.assertEqual(results[1][3], u'my very great pleasure')

        tokens_extractor = EuroparlPoSExtractor('en', ['nl'], tokens=[('w1.19', 'w1.17')])
        generator = tokens_extractor.generate_results(os.path.join(EUROPARL_DATA, 'en'))
        self.assertRaises(ValueError, next, generator)

    def test_metadata(self):
        metadata_extractor = EuroparlPoSExtractor('en', [], lemmata=['when'],
                                                  metadata=[('topic', 'text'), ('damsl_act_tag', 's')])
        results = self.merge_results(metadata_extractor.generate_results(os.path.join(SWITCHBOARD_DATA, 'en')))
        self.assertEqual(len(results), 5)
        self.assertEqual(results[0][6], u'CHILD CARE')
        self.assertEqual(results[0][7], u'sd')
        self.assertEqual(results[4][7], u'qy')

    def test_since(self):
        metadata_extractor = EuroparlSinceDurationExtractor('nl', [])
        results = self.merge_results(metadata_extractor.generate_results(os.path.join(EUROPARL_DATA, 'nl')))
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0][3], u'sinds tien jaar')

    def test_articles(self):
        article_extractor = EuroparlFrenchArticleExtractor('fr', [])
        results = self.merge_results(article_extractor.generate_results(os.path.join(EUROPARL_DATA, 'fr')))
        self.assertEqual(len(results), 2041)
        self.assertEqual(results[0][2], u'indefinite partitive')

    def test_past_perfect(self):
        past_perfect_extractor = EuroparlPerfectExtractor('en', [], tense=PAST)
        results = self.merge_results(past_perfect_extractor.generate_results(os.path.join(EUROPARL_DATA, 'en')))
        self.assertEqual(len(results), 5)
        self.assertEqual(results[0][3], u'had preordained')

        past_perfect_extractor = EuroparlPerfectExtractor('nl', [], tense=PAST)
        results = self.merge_results(past_perfect_extractor.generate_results(os.path.join(EUROPARL_DATA, 'nl')))
        self.assertEqual(len(results), 5)
        self.assertEqual(results[0][3], u'had gelegd')

        past_perfect_extractor = EuroparlPerfectExtractor('fr', [], tense=PAST)
        results = self.merge_results(past_perfect_extractor.generate_results(os.path.join(EUROPARL_DATA, 'fr')))
        self.assertEqual(len(results), 9)
        self.assertEqual(results[0][3], u'avait rétabli')

        past_perfect_extractor = EuroparlPerfectExtractor('de', [], tense=PAST)
        results = self.merge_results(past_perfect_extractor.generate_results(os.path.join(DCEP_DATA, 'de')))
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0][3], u'hatte beschlossen')

        past_perfect_extractor = EuroparlPerfectExtractor('sv', [], tense=PAST)
        results = self.merge_results(past_perfect_extractor.generate_results(os.path.join(EUROPARL_DATA, 'sv')))
        self.assertEqual(len(results), 7)
        self.assertEqual(results[0][3], u'hade kunnat')

    def test_languages(self):
        sv_extractor = EuroparlPerfectExtractor('sv', [])
        results = self.merge_results(sv_extractor.generate_results(os.path.join(EUROPARL_DATA, 'sv')))
        self.assertEqual(len(results), 38)
        self.assertEqual(results[0][3], u'har gjort')



import glob
import os

BNC_CONFIG = os.path.join(os.path.dirname(__file__), 'bnc.cfg')


class BaseBNC(object):
    def get_config(self):
        return BNC_CONFIG

    def list_filenames(self, dir_name):
        return sorted(glob.glob(os.path.join(dir_name, '*.xml')))

    def get_genre(self, tree):
        return tree.xpath('.//classCode')[0].text



MARKUP = u'**{}**'


class Word(object):
    """
    Each Word consists of a word, its lemma, and a designation if this is a verb.
    """
    def __init__(self, word, lemma, is_verb, xml_id):
        self.word = word.strip() if word else ' '
        self.lemma = lemma.strip() if lemma else '?'
        self.is_verb = is_verb
        self.xml_id = xml_id


class MultiWordExpression(object):
    def __init__(self, xml_sentence=None):
        self.xml_sentence = xml_sentence
        self.words = []

    def add_word(self, word, lemma, is_verb, xml_id):
        """
        Adds a word to the MultiWordExpression.
        """
        self.words.append(Word(word, lemma, is_verb, xml_id))

    def prepend_word(self, word, lemma, is_verb, xml_id):
        """
        Prepends a word to the MultiWordExpression.
        """
        self.words.insert(0, Word(word, lemma, is_verb, xml_id))

    def verbs(self):
        """
        Extracts the verbs from the MultiWordExpression.
        """
        return [w.word for w in self.words if w.is_verb]

    def verbs_to_string(self):
        """
        Returns the verbs from the MultiWordExpression as a string.
        """
        return ' '.join(self.verbs())

    def verb_ids(self):
        return ' '.join([w.xml_id for w in self.words if w.is_verb])

    def words_between(self):
        """
        Returns the total number of non-verbs in a MultiWordExpression.
        """
        return len([w.word for w in self.words if not w.is_verb])

    def words_between_verbs(self):
        """
        Returns the number of non-verbs in a MultiWordExpression between the verbs.
        """
        result = []
        current_count = 0
        for w in self.words:
            if w.is_verb:
                result.append(current_count)
                current_count = 0
            else:
                current_count += 1

        return result

    def get_sentence_words(self):
        s = []
        # TODO: this xPath-expression might be specific for a corpus
        for w in self.xml_sentence.xpath('.//w'):
            s.append(w.text.strip() if w.text else ' ')
        return ' '.join(s)

    def get_sentence_id(self):
        return self.xml_sentence.get('id')

    def mark_sentence(self):
        """
        Marks the MultiWordExpression in a full sentence.
        """
        # TODO: this doesn't work if no xml_sentence is given
        if self.xml_sentence is None:
            return ''

        # TODO: this is a bit iffy, another idea could be to compose the sentence from the remaining siblings
        # To find the pp in the full text, simply join all the parts of the pp
        pp_text = ' '.join([w.word for w in self.words])

        # For the replacement, mark the verbs with the MARKUP
        if len(self.words) == len(self.verbs()):
            marked_pp = MARKUP.format(pp_text)
        else:
            marked_pp = ' '.join([MARKUP.format(w.word) if w.is_verb else w.word for w in self.words])

        return self.get_sentence_words().replace(pp_text, marked_pp)


class PresentPerfect(MultiWordExpression):
    """
    A present perfect consists of a list of Words.
    """

    def __init__(self, aux_verb, aux_lemma, xml_id, xml_sentence=None):
        """
        A present perfect is initiated by an auxiliary verb.
        """
        super(PresentPerfect, self).__init__(xml_sentence)
        self.is_passive = False
        self.is_continuous = False
        self.is_reflexive = False
        self.add_word(aux_verb, aux_lemma, True, xml_id)

    def extend(self, present_perfect):
        """
        Extends a present perfect with another one (used to create a present perfect passive or continuous).
        """
        for i, w in enumerate(present_perfect.words):
            if i == 0:
                continue
            self.add_word(w.word, w.lemma, w.is_verb, w.xml_id)

        self.is_passive = not present_perfect.is_continuous
        self.is_continuous = present_perfect.is_continuous

    def perfect_lemma(self):
        """
        Returns the lemma of the perfect: the lemma of the last word of the construction.
        """
        return self.words[-1].lemma

    def perfect_type(self):
        result = 'present perfect'
        if self.is_passive:
            result += ' passive'
        if self.is_continuous:
            result += ' continuous'
        return result


class Alignment(object):
    def __init__(self, sources, targets, certainty=None):
        self.sources = sources
        self.targets = targets
        self.certainty = certainty



import codecs
import csv

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

# Output formats
TXT = 'txt'
XML = 'xml'


def get_adjacent_line_number(segment_number, i):
    """
    Returns the next segment number + i.
    Segment numbers have the form "pN.sM", where N/M are positive integers.
    """
    split = segment_number.split('s')
    adj = int(split[1]) + i
    return split[0] + 's' + str(adj)


class UnicodeWriter:
    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    Copied from https://docs.python.org/2/library/csv.html#examples
    """

    def __init__(self, f, dialect=csv.excel, encoding='utf-8', **kwds):
        # Redirect output to a queue
        self.queue = StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwds)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writerow(self, row):
        self.writer.writerow([s.encode('utf-8') for s in row])
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        data = data.decode('utf-8')
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)



from abc import ABCMeta, abstractmethod
import os

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser

from .base import BaseExtractor

LEMMATA_CONFIG = os.path.join(os.path.dirname(__file__), '../config/{language}_lemmata.txt')


class PoSExtractor(BaseExtractor):
    __metaclass__ = ABCMeta

    def __init__(self, language_from, languages_to=None, pos=None, regex=None, **kwargs):
        """
        Initializes the extractor for the given source and target language(s).
        Reads in the config for the source language.
        :param language_from: the source language
        :param languages_to: the target language(s)
        :param pos: A list of part-of-speech tags
        :param regex: A list of regular expressions
        """
        super(PoSExtractor, self).__init__(language_from, languages_to, **kwargs)

        self.pos = pos
        self.regex = regex

    @abstractmethod
    def preprocess_found(self, word):
        """
        Preprocesses the found word: potentially add more words to the found words, or filter based on lemmata.
        """
        raise NotImplementedError

    def get_type(self, words):
        """
        Return the type for the found word(s). A sensible default is the part-of-speech of the first found word.
        """
        return words[0].get(self.config.get(self.l_from, 'pos'))



# -*- encoding: utf-8 -*-

from abc import ABCMeta, abstractmethod
import codecs
import string
import os

from .base import BaseExtractor
from .models import PresentPerfect
from .wiktionary import get_translations

# List of verbs that have BE instead of HAVE as their auxiliary
AUX_BE_CONFIG = os.path.join(os.path.dirname(__file__), 'config/{language}_aux_be.txt')

# Tenses
PRESENT = 'present'
PAST = 'past'
# FUTURE = 'future'  # TODO: implement this somewhere in the near future


class PerfectExtractor(BaseExtractor):
    __metaclass__ = ABCMeta

    def __init__(self, language_from, languages_to=None, search_in_to=False, tense=PRESENT, **kwargs):
        """
        Initializes the PerfectExtractor for the given source and target language(s).
        Reads in the config for the source language,
        as well as the list of verbs that use 'to be' as auxiliary verb for both source and target language(s).
        :param language_from: the source language
        :param languages_to: the target language(s)
        :param search_in_to: whether to look for perfects in the target language
        :param tense: whether to search for present, past or future perfects
        """
        super(PerfectExtractor, self).__init__(language_from, languages_to, **kwargs)

        self.search_in_to = search_in_to
        self.tense = tense

        # Read the list of verbs that use 'to be' as auxiliary verb per language
        self.aux_be_list = {}
        languages = [self.l_from]
        if search_in_to:
            languages.extend(self.l_to)
        for language in languages:
            aux_be_list = []
            if self.config.get(language, 'lexical_bound'):
                with codecs.open(AUX_BE_CONFIG.format(language=language), 'r', 'utf-8') as lexicon:
                    aux_be_list = lexicon.read().split()
            self.aux_be_list[language] = aux_be_list

    @abstractmethod
    def get_line_and_pp(self, tree, language_to, segment_number):
        """
        Returns the full line for a segment number, as well as the PresentPerfect found (or None if none found).
        TODO: handle more than one here? => bug
        """
        raise NotImplementedError

    def is_lexically_bound(self, language, pp, aux_verb, past_participle, w_before=None):
        """
        Checks if the perfect is lexically bound to the auxiliary verb.
        If not, we are not dealing with a present perfect here.
        """
        lemma_attr = self.config.get('all', 'lemma_attr')
        aux_be = self.config.get(language, 'lexical_bound')

        # If lexical bounds do not exist or we're dealing with an auxiliary verb that is unbound, return True
        # Note: we check with "not in", because in French the lemma can be e.g. 'suivre|être'
        if not aux_be or aux_be not in aux_verb.get(lemma_attr):
            return True

        # Else, check if we are dealing with a reflexive present perfect (in that case, there is no lexical bound)
        if self.is_reflexive(language, w_before):
            pp.prepend_word(w_before[0].text, w_before[0].get(lemma_attr), True, w_before[0].get('id'))
            pp.is_reflexive = True
            return True

        # Finally, check whether the past participle is in the list of bound verbs
        return past_participle.get(lemma_attr) in self.aux_be_list[language]

    def is_reflexive(self, language, w_before):
        lemma_attr = self.config.get('all', 'lemma_attr')
        reflexive_lemmata = self.config.get(language, 'reflexive_lemmata').split(',')

        precondition = reflexive_lemmata and w_before is not None and len(w_before) >= 2
        if precondition:
            prev_reflexive = w_before[0].get(lemma_attr) in reflexive_lemmata
            # TODO: below condition is language-specific, and does not yet work for e.g. negation
            prevprev_pronoun = w_before[0].get(lemma_attr) not in ['nous', 'vous'] or self.get_pos(language, w_before[1]) == 'PRO:PER'
            return prev_reflexive and prevprev_pronoun
        else:
            return False

    def check_present_perfect(self, auxiliary, language, check_ppp=True, check_ppc=False, check_preceding=False):
        """
        Checks whether this element (i.e. the auxiliary) is the start of a present perfect (pp),
        a present perfect continuous (ppc) or passive present perfect (ppp).
        If it is, the complete construction is returned as a PresentPerfect object.
        If not, None is returned.
        """
        lemma_attr = self.config.get('all', 'lemma_attr')
        perfect_tags = self.config.get(language, 'perfect_tags').split(',')
        check_ppp = check_ppp and self.config.getboolean(language, 'ppp')
        ppp_lemma = self.config.get(language, 'ppp_lemma')
        check_ppc = check_ppc and self.config.getboolean(language, 'ppc')
        ppc_tags = self.config.get(language, 'ppc_tags').split(',')
        stop_tags = tuple(self.config.get(language, 'stop_tags').split(','))
        allow_reversed = self.config.getboolean(language, 'allow_reversed')

        # Retrieves the auxiliaries, or a fallback if there are none provided
        aux_fallback = 'aux_words'
        aux = aux_fallback + ('_{}'.format(self.tense) if self.tense != PRESENT else '')
        l_config = dict(self.config.items(language))  # In Python 3, we could consider self.config[language] instead.
        aux_words = l_config.get(aux, l_config.get(aux_fallback)).split(',')

        # Start a potential present perfect
        s = self.get_sentence(auxiliary)
        pp = PresentPerfect(auxiliary.text, auxiliary.get(lemma_attr), auxiliary.get('id'), s)
        is_pp = False

        # Check if the starting auxiliary is actually allowed
        if any(aux_words) and auxiliary.text.lower() not in aux_words:
            return None

        # Loop over the siblings of the current element.
        siblings = self.get_siblings(auxiliary, s.get('id'), check_preceding)
        for n, sibling in enumerate(siblings):
            # If the tag of the sibling is the perfect tag, we found a present perfect!
            sibling_pos = self.get_pos(language, sibling)
            sibling_lemma = sibling.get(lemma_attr)
            if sibling_pos in perfect_tags:
                # Check if the sibling is lexically bound to the auxiliary verb
                # (only if we're not checking for passive present perfect)
                before = self.get_siblings(auxiliary, s.get('id'), not check_preceding)
                if check_ppp and not self.is_lexically_bound(language, pp, auxiliary, sibling, before):
                    break

                # Check if the lemma is not in the lemmata list, if so break, unless we found a potential ppp
                if not self.in_lemmata_list(sibling_lemma):
                    if not(check_ppp and sibling_lemma == ppp_lemma):
                        break

                pp.add_word(sibling.text, sibling_lemma, True, sibling.get('id'))
                is_pp = True

                # ... now check whether this is a passive present perfect or present perfect continuous (by recursion)
                if check_ppp and sibling_lemma == ppp_lemma:
                    ppp = self.check_present_perfect(sibling,
                                                     language,
                                                     check_ppp=False,
                                                     check_ppc=True,
                                                     check_preceding=check_preceding)
                    if ppp:
                        pp.extend(ppp)
                    elif not self.in_lemmata_list(sibling_lemma):
                        is_pp = False
                break
            # Check if this is a present perfect continuous (in the recursion step)
            elif check_ppc and sibling_pos in ppc_tags and self.in_lemmata_list(sibling_lemma):
                pp.add_word(sibling.text, sibling_lemma, True, sibling.get('id'))
                pp.is_continuous = True
                is_pp = True
                break
            # Stop looking at punctuation or stop tags
            elif (sibling.text and sibling.text in string.punctuation) or (sibling_pos and sibling_pos.startswith(stop_tags)):
                break
            # We didn't break yet? Then add this sibling as a potential non-verb part of the present perfect.
            else:
                pp.add_word(sibling.text, sibling_lemma, False, sibling.get('id'))

        # If we haven't yet found a past participle, and we are allowed to look in the other direction,
        # try to find a past participle by looking backwards in the sentence.
        if not is_pp and allow_reversed and not check_preceding:
            pp = self.check_present_perfect(auxiliary, language, check_ppp=check_ppp, check_preceding=True)
            if pp:
                is_pp = True

        return pp if is_pp else None

    def in_lemmata_list(self, element):
        """
        Returns whether the given element is in the lemmata list.
        Returns True when there is no lemmata list given.
        """
        return not self.lemmata_list or element in self.lemmata_list

    def find_translated_present_perfects(self, translated_tree, language_to, translated_lines):
        """
        Finds present perfects in translated sentences.
        """
        translated_pps = []
        translated_sentences = []
        translated_marked_sentences = []

        if translated_lines and any(translated_lines):
            for t in translated_lines:
                sentence, translation, translated_pp = self.get_line_and_pp(translated_tree, language_to, t)
                translated_pps.append(translated_pp)
                translated_sentences.append(sentence)
                translated_marked_sentences.append(translation)

        return translated_pps, translated_sentences, translated_marked_sentences

    def check_translated_pps(self, pp, translated_present_perfects, language_to):
        """
        Checks whether the translated present perfects found form an actual translation of the present perfect.
        """
        results = []
        for tpp in translated_present_perfects:
            if tpp:
                translations = get_translations(pp.perfect_lemma(), self.l_from, language_to)
                if tpp.perfect_lemma() in translations:
                    results.append('yes')
                else:
                    results.append('unknown')
        return results



# -*- coding: utf-8 -*-

import os
import unittest

from lxml import etree

from perfectextractor.apps.extractor.models import PresentPerfect
from perfectextractor.corpora.dpc.extractor import DPCPerfectExtractor

DATA_FOLDER = os.path.join(os.path.dirname(__file__), 'data/dpc')


class TestDPCExtractor(unittest.TestCase):
    def setUp(self):
        self.en_extractor = DPCPerfectExtractor('en', ['nl', 'fr'], search_in_to=True)
        self.nl_extractor = DPCPerfectExtractor('nl', ['en', 'fr'], search_in_to=True)
        self.fr_extractor = DPCPerfectExtractor('fr', ['en', 'nl'], search_in_to=True)

        self.document = os.path.join(DATA_FOLDER, 'dpc-bmm-001071-')
        align_fr = etree.parse(os.path.join(DATA_FOLDER, 'dpc-bmm-001071-nl-fr-tei.xml'))
        align_en = etree.parse(os.path.join(DATA_FOLDER, 'dpc-bmm-001071-nl-en-tei.xml'))
        self.alignmenttrees = {'en': align_en, 'fr': align_fr}

    def test_init(self):
        self.assertEqual(self.en_extractor.config.get('en', 'perfect_tags'), 'VBN')
        self.assertIn('devenir', self.en_extractor.aux_be_list['fr'])

    def test_get_translated_lines(self):
        lines, alignment = self.en_extractor.get_translated_lines(self.alignmenttrees, 'en', 'nl', 'p1.s3')
        self.assertEqual(lines, set(['p1.s3']))
        self.assertEqual(alignment, '2=>1')
        lines, alignment = self.en_extractor.get_translated_lines(self.alignmenttrees, 'nl', 'en', 'p1.s3')
        self.assertEqual(lines, set(['p1.s3', 'p1.s4']))
        self.assertEqual(alignment, '1=>2')
        lines, alignment = self.en_extractor.get_translated_lines(self.alignmenttrees, 'fr', 'nl', 'p1.s13')
        self.assertEqual(lines, set(['p1.s12']))
        self.assertEqual(alignment, '1=>1')
        lines, alignment = self.en_extractor.get_translated_lines(self.alignmenttrees, 'nl', 'fr', 'p1.s13')
        self.assertEqual(lines, set(['p1.s14']))
        self.assertEqual(alignment, '2=>1')
        lines, alignment = self.en_extractor.get_translated_lines(self.alignmenttrees, 'en', 'fr', 'p1.s16')
        self.assertEqual(lines, set(['p1.s16']))
        self.assertEqual(alignment, '1=>1')
        lines, alignment = self.en_extractor.get_translated_lines(self.alignmenttrees, 'fr', 'en', 'p1.s4')
        self.assertEqual(lines, set(['p1.s3', 'p1.s4']))
        self.assertEqual(alignment, '2=>2')

    def test_get_line_by_number(self):
        tree = etree.parse(self.document + 'en-tei.xml')
        line = self.en_extractor.get_line_and_pp(tree, 'en', 'p1.s16')
        pp = PresentPerfect('have', 'have', 'test_id')
        pp.add_word('attained', 'attain', True, 'test_id')
        # self.assertIn(u'**have attained**', line[0])
        self.assertEqual(pp.verbs_to_string(), line[1].verbs_to_string())

    def test_get_original_language(self):
        orig_lang = self.en_extractor.get_original_language(self.document)
        self.assertEqual(orig_lang, 'unknown')

    def test_check_present_perfect(self):
        pass

    def test_find_translated_present_perfects(self):
        pass

    def test_check_translated_pps(self):
        pass

    def test_process_file(self):
        pass
        #print self.en_extractor.process_file('data/dpc-bmm-001071-en-tei.xml')



import setuptools


def readme():
    with open('README.md') as f:
        return f.read()


setuptools.setup(
    name='perfectextractor',
    version='0.3.1',
    author='Martijn van der Klis',
    author_email='M.H.vanderKlis@uu.nl',
    description='Extracting present perfects (and related forms) from parallel corpora',
    long_description=readme(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Topic :: Text Processing :: Linguistic',
    ],
    url='https://github.com/UUDigitalHumanitieslab/perfectextractor',
    license='MIT',
    packages=setuptools.find_packages(),
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=[
      'click',
      'lxml',
      'requests',
    ],
)



# -*- encoding: utf-8 -*-

import os
import time

import click
from lxml import etree

from perfectextractor.apps.extractor.base import BaseExtractor
from perfectextractor.apps.extractor.models import Alignment
from perfectextractor.apps.extractor.perfectextractor import PerfectExtractor, PRESENT
from perfectextractor.apps.extractor.posextractor import PoSExtractor
from perfectextractor.apps.extractor.recentpastextractor import RecentPastExtractor
from perfectextractor.apps.extractor.xml_utils import get_sentence_from_element
from perfectextractor.apps.extractor.utils import XML

from .base import BaseEuroparl


class EuroparlExtractor(BaseEuroparl, BaseExtractor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._index = dict()  # save segments indexed by id

    def process_file(self, filename):
        """
        Processes a single file.
        """
        t0 = time.time()
        click.echo('Now processing {}...'.format(filename))

        # Parse the current tree (create a iterator over 's' elements)
        s_trees = etree.iterparse(filename, tag='s')

        # Parse the alignment and translation trees
        alignment_trees, translation_trees = self.parse_alignment_trees(filename)

        t1 = time.time()
        click.echo('Finished parsing trees, took {:.3} seconds'.format(t1 - t0))

        # Fetch the results
        results = self.fetch_results(filename, s_trees, alignment_trees, translation_trees)

        click.echo('Finished fetching results, took {:.3} seconds'.format(time.time() - t1))

        self._index = dict()  # free index memory
        return results

    def fetch_results(self, filename, s_trees, alignment_trees, translation_trees):
        """
        Fetches the results for a file
        :param filename: The current filename
        :param s_trees: The XML trees for all 's' elements in the file
        :param alignment_trees: The alignment XML trees, per target language
        :param translation_trees: The translation XML trees, per target language
        :return: All results in a list
        """
        results = list()
        # Loop over all sentences
        for _, s in s_trees:
            if self.sentence_ids and s.get('id') not in self.sentence_ids:
                continue

            result = list()
            result.append(os.path.basename(filename))
            result.append(s.get('id'))
            result.append('')
            result.append('')
            result.append('')
            if self.output == XML:
                result.append('<root>' + etree.tostring(s, encoding=str) + '</root>')
            else:
                result.append(self.get_sentence_words(s))
            self.append_metadata(None, s, result)

            for language_to in self.l_to:
                if language_to in translation_trees:
                    # TODO: deal with source_lines
                    source_lines, translated_lines, alignment_type = self.get_translated_lines(alignment_trees,
                                                                                               self.l_from,
                                                                                               language_to,
                                                                                               s.get('id'))
                    result.append(alignment_type)
                    if self.output == XML:
                        translated_sentences = [self.get_line(translation_trees[language_to], line) for line in translated_lines]
                        result.append('<root>' + '\n'.join(translated_sentences) + '</root>' if translated_sentences else '')
                    else:
                        translated_sentences = [self.get_line_as_xml(translation_trees[language_to], line) for line in translated_lines]
                        result.append('\n'.join([self.get_sentence_words(ts) for ts in translated_sentences]) if translated_sentences else '')
                else:
                    # If no translation is available, add empty columns
                    result.extend([''] * 2)

            results.append(result)
        return results

    def get_sentence_words(self, sentence):
        # TODO: this is copied from apps/models.py. Consider refactoring!
        s = []
        # TODO: this xPath-expression might be specific for a corpus
        for w in sentence.xpath('.//w'):
            s.append(w.text.strip() if w.text else ' ')
        return ' '.join(s)

    def get_line_by_number(self, tree, segment_number):
        """
        Returns the full line for a segment number.
        TODO: handle more than one here? => bug
        """
        result = None

        line = tree.xpath('//s[@id="' + segment_number + '"]')
        if line is not None:
            result = line[0]

        return result

    def get_sentence(self, element):
        return element.xpath('ancestor::s')[0]

    def get_siblings(self, element, sentence_id, check_preceding):
        siblings = element.xpath('ancestor::s//w')
        if check_preceding:
            siblings = siblings[:siblings.index(element)]
            siblings = siblings[::-1]
        else:
            siblings = siblings[siblings.index(element) + 1:]
        return siblings

    def _segment_by_id(self, tree, id):
        if tree not in self._index:
            self._index[tree] = dict()
            for segment in tree.xpath('//s'):
                self._index[tree][segment.get('id')] = segment
        return self._index[tree].get(id)

    def get_line_as_xml(self, tree, segment_number):
        return self._segment_by_id(tree, segment_number)

    def get_line(self, tree, segment_number):
        line = self._segment_by_id(tree, segment_number)
        if line is not None:
            return etree.tostring(line, encoding=str)
        else:
            return None

    def get_translated_lines(self, alignment_trees, language_from, language_to, segment_number):
        """
        Returns the translated segment numbers (could be multiple) for a segment number in the original text.

        Alignment lines in the Europarl corpus look like this:
            <linkGrp targType="s" fromDoc="en/ep-00-01-17.xml.gz" toDoc="nl/ep-00-01-17.xml.gz">
                <link xtargets="1;1" />
                <link xtargets="2;2 3" />
                <link xtargets="3;4 5" />
            </linkGrp>

        To get from language A to B, we should order the languages

        This function supports n-to-n alignments, as it will return both the source and translated lines as a list.
        """
        from_lines = []
        to_lines = []
        certainty = None

        sl = sorted([language_from, language_to])
        for alignment in alignment_trees[language_to]:
            if sl[0] == language_from:
                if segment_number in alignment.sources:
                    from_lines = alignment.sources
                    to_lines = alignment.targets
                    certainty = alignment.certainty
                    break
            else:
                if segment_number in alignment.targets:
                    from_lines = alignment.targets
                    to_lines = alignment.sources
                    certainty = alignment.certainty
                    break

        if not any(to_lines):
            to_lines = []

        alignment_str = '{} => {}'.format(len(from_lines), len(to_lines)) if to_lines else ''

        return from_lines, to_lines, alignment_str

    def parse_alignment_trees(self, filename, include_translations=True):
        data_folder = os.path.dirname(os.path.dirname(filename))

        # Cache the alignment XMLs on the first run
        if not self.alignment_xmls:
            for language_to in self.l_to:
                sl = sorted([self.l_from, language_to])
                alignment_file = os.path.join(data_folder, '-'.join(sl) + '.xml')
                if os.path.isfile(alignment_file):
                    alignment_tree = etree.parse(alignment_file)
                    self.alignment_xmls[language_to] = alignment_tree
                elif include_translations:
                    click.echo('No alignment file found for {} to {}'.format(filename, language_to))

        alignment_trees = dict()
        translation_trees = dict()
        for language_to in self.alignment_xmls.keys():
            sl = sorted([self.l_from, language_to])
            alignment_tree = self.alignment_xmls[language_to]
            base_filename = os.path.basename(filename)
            doc = '{}/{}'.format(self.l_from, base_filename)
            doc_gz = doc + '.gz'  # Europarl uses .gz natively, deal with both options
            path = '@fromDoc="{}"' if sl[0] == self.l_from else '@toDoc="{}"'
            linkGrps = alignment_tree.xpath('//linkGrp[{} or {}]'.format(path.format(doc), path.format(doc_gz)))

            if not linkGrps:
                if include_translations:
                    click.echo('No translation found for {} to {}'.format(filename, language_to))
            elif len(linkGrps) == 1:
                linkGrp = linkGrps[0]

                if include_translations:
                    translation_link = linkGrp.get('toDoc') if sl[0] == self.l_from else linkGrp.get('fromDoc')
                    if translation_link.endswith('.gz'):   # See comment above: Europarl uses .gz as extension
                        translation_link = translation_link[:-3]
                    translation_file = os.path.join(data_folder, translation_link)
                    translation_trees[language_to] = etree.parse(translation_file)

                alignments = []
                for link in linkGrp.xpath('./link'):
                    xtargets = link.get('xtargets').split(';')
                    sources = xtargets[0].split(' ')
                    targets = xtargets[1].split(' ')
                    certainty = link.get('certainty', None)
                    alignments.append(Alignment(sources, targets, certainty))

                alignment_trees[language_to] = alignments
            else:
                click.echo('Multiple translations found for {} to {}'.format(filename, language_to))

        return alignment_trees, translation_trees

    def average_alignment_certainty(self, alignment_trees):
        certainties_sum = 0
        certainties_len = 0
        for language, alignments in alignment_trees.items():
            certainties = [float(a.certainty) if a.certainty else 0 for a in alignments]
            certainties_sum += sum(certainties)
            certainties_len += len(certainties)

        return certainties_sum / float(certainties_len) if certainties_len > 0 else 0

    def sort_by_alignment_certainty(self, file_names):
        certainties = dict()
        for file_name in file_names:
            alignment_trees, _ = self.parse_alignment_trees(file_name, include_translations=False)
            if len(alignment_trees) == len(self.l_to):  # Only add files that have translations in all languages
                certainties[file_name] = self.average_alignment_certainty(alignment_trees)

        sorted_by_value = sorted(certainties.items(), key=lambda kv: kv[1], reverse=True)
        return [f[0] for f in sorted_by_value]

    def filter_by_file_size(self, file_names):
        results = []
        for file_name in file_names:
            file_size = etree.parse(file_name).xpath('count(//s)')
            if self.min_file_size <= file_size <= self.max_file_size:
                results.append(file_name)

        return results

    def list_directories(self, path):
        return filter(lambda x: x.endswith(self.l_from), super(EuroparlExtractor, self).list_directories(path))


class EuroparlPoSExtractor(EuroparlExtractor, PoSExtractor):
    def fetch_results(self, filename, s_trees, alignment_trees, translation_trees):
        """
        Processes a single file.
        """
        results = []

        # Find potential words matching the part-of-speech
        pos_attr = self.config.get(self.l_from, 'pos')
        lemma_attr = self.config.get('all', 'lemma_attr')
        xp = './/w[contains(" {value} ", concat(" ", @{element}, " "))]'
        ns = {}
        if self.pos:
            xpath = xp.format(element=pos_attr, value=' '.join(self.pos))
        elif self.tokens:
            xpath = xp.format(element='id', value=' '.join(self.tokens.keys()))
        elif self.regex:
            # prepare a pattern that combines multiple regexps using OR operators
            # and non-capturing groups
            pattern = '|'.join('(?:{})'.format(r) for r in self.regex)

            # special namespace required for enabling regular expresssion functions
            ns = {"re": "http://exslt.org/regular-expressions"}
            xpath = './/w[re:test(., "{pattern}", "i")]'.format(pattern=pattern)
        else:
            xpath = xp.format(element=lemma_attr, value=' '.join(self.lemmata_list))

        for _, s in s_trees:
            if self.sentence_ids and s.get('id') not in self.sentence_ids:
                continue

            for w in s.xpath(xpath, namespaces=ns):
                words = self.preprocess_found(w)

                if not words:
                    continue

                result = list()
                result.append(os.path.basename(filename))
                result.append(s.get('id'))
                result.append(self.get_type(words))
                result.append(' '.join([w.text for w in words]))
                result.append(' '.join([w.get('id') for w in words]))
                if self.output == XML:
                    result.append('<root>' + etree.tostring(s, encoding=str) + '</root>')
                else:
                    result.append(self.get_sentence_words(s))
                self.append_metadata(w, s, result)

                # Find the translated lines
                for language_to in self.l_to:
                    if language_to in translation_trees:
                        # TODO: deal with source_lines
                        source_lines, translated_lines, alignment_type = self.get_translated_lines(alignment_trees,
                                                                                                   self.l_from,
                                                                                                   language_to,
                                                                                                   s.get('id'))
                        result.append(alignment_type)
                        if self.output == XML:
                            translated_sentences = [self.get_line(translation_trees[language_to], line) for line in translated_lines]
                            result.append('<root>' + '\n'.join(translated_sentences) + '</root>' if translated_sentences else '')
                        else:
                            translated_sentences = [self.get_line_as_xml(translation_trees[language_to], line) for line in translated_lines]
                            result.append('\n'.join([self.get_sentence_words(ts) for ts in translated_sentences]) if translated_sentences else '')
                    else:
                        # If no translation is available, add empty columns
                        result.extend([''] * 2)

                results.append(result)

        return results

    def preprocess_found(self, word):
        """
        Preprocesses the found word:
        - removes a word if it does not occur in the lemmata list
        - removes a word if it is not in the correct position in the sentence
        Returns the found word as a list, as it might be interesting to include words before and after
        (see e.g. EuroparlFrenchArticleExtractor)
        """
        result = [word]

        lemma_attr = self.config.get('all', 'lemma_attr')
        if self.lemmata_list and word.get(lemma_attr) not in self.lemmata_list:
            result = []

        if self.position and not word.get('id').endswith('.' + str(self.position)):
            result = []

        if self.tokens:
            end_token = self.tokens.get(word.get('id'))
            next_word = word.getnext()
            if next_word is None:
                raise ValueError('End token {} not found'.format(end_token))
            else:
                while next_word.get('id') != end_token:
                    result.append(next_word)

                    next_word = next_word.getnext()
                    if next_word is None:
                        raise ValueError('End token {} not found'.format(end_token))
                else:
                    result.append(next_word)

        return result


class EuroparlFrenchArticleExtractor(EuroparlPoSExtractor):
    def __init__(self, language_from, languages_to):
        """
        Initializes the EuroparlFrenchArticleExtractor with a set of part-of-speeches and lemmata that the found
        words should adhere to. Also initializes a list of particles that could appear before a determiner.
        :param language_from: The language to find the specified part-of-speeches in.
        :param languages_to: The languages to extract the aligned sentences from.
        """
        super(EuroparlFrenchArticleExtractor, self).__init__(language_from, languages_to, ['DET:ART', 'PRP:det'])

        self.lemmata_list = ['le', 'un', 'du']
        self.particles = ['de', 'du']

    def preprocess_found(self, word):
        """
        Removes a word if does not occur in the lemmata list, and then checks if it might be preceded by a particle.
        If so, add the particle to the found words.
        """
        result = []

        lemma_attr = self.config.get('all', 'lemma_attr')
        for w in super(EuroparlFrenchArticleExtractor, self).preprocess_found(word):
            prev = w.getprevious()
            if prev is not None and prev.get(lemma_attr) in self.particles:
                result.append(prev)

            result.append(word)

        return result

    def get_type(self, words):
        """
        Return the type for a found article: definite, indefinite or partitive.
        For 'des', this is quite hard to decide, so we leave both options open.
        """
        result = ''
        lemma_attr = self.config.get('all', 'lemma_attr')

        if words[-1].get(lemma_attr) == 'le':
            result = 'definite'
        elif words[-1].get(lemma_attr) == 'un':
            result = 'indefinite'

        if words[0].get(lemma_attr) in self.particles:
            if result:
                result += ' '
            result += 'partitive'

        if words[0].text == 'des':
            result = 'indefinite/partitive'

        return result


class EuroparlSinceDurationExtractor(EuroparlPoSExtractor):
    def __init__(self, language_from, languages_to, **kwargs):
        """
        Initializes the EuroparlSinceDurationExtractor with a set of lemmata, part-of-speech for numbers,
        and time units, so that it should find [seit] + [number] + [unit of time].
        :param language_from: The language to find the specified part-of-speeches in.
        :param languages_to: The languages to extract the aligned sentences from.
        """
        super(EuroparlPoSExtractor, self).__init__(language_from, languages_to, **kwargs)

        self.lemmata_list = self.config.get(self.l_from, 'since_lem').split(',')
        self.number_pos = self.config.get(self.l_from, 'since_number_pos').split(',')
        self.time_units = self.config.get(self.l_from, 'since_time_units').split(',')

    def preprocess_found(self, word):
        """
        Removes a word if does not occur in the lemmata list, and then checks if it might be preceded by a particle.
        If so, add the particle to the found words.
        """
        result = []

        lemma_attr = self.config.get('all', 'lemma_attr')
        pos_attr = self.config.get(self.l_from, 'pos')

        for w in super(EuroparlSinceDurationExtractor, self).preprocess_found(word):
            result.append(word)

            next_word = w.getnext()
            if next_word is not None and next_word.get(pos_attr) in self.number_pos:
                result.append(next_word)
            else:
                result = None
                continue

            next_word2 = next_word.getnext()
            if next_word2 is not None and next_word2.get(lemma_attr) in self.time_units:
                result.append(next_word2)
            else:
                result = None
                continue

        return result


class EuroparlRecentPastExtractor(EuroparlExtractor, RecentPastExtractor):
    def fetch_results(self, filename, s_trees, alignment_trees, translation_trees):
        """
        Processes a single file.
        """
        results = []
        # Find potential recent pasts (per sentence)
        for _, s in s_trees:
            if self.sentence_ids and s.get('id') not in self.sentence_ids:
                continue

            for w in s.xpath(self.config.get(self.l_from, 'rp_xpath')):
                rp = self.check_recent_past(w, self.l_from)

                if rp:
                    result = list()
                    result.append(os.path.basename(filename))
                    result.append(s.get('id'))
                    result.append(u'passé récent')
                    result.append(rp.verbs_to_string())
                    result.append(rp.verb_ids())
                    result.append('<root>' + etree.tostring(rp.xml_sentence, encoding=str) + '</root>')
                    self.append_metadata(w, s, result)

                    found_trans = False
                    for language_to in self.l_to:
                        if language_to in translation_trees:
                            # TODO: deal with source_lines
                            source_lines, translated_lines, alignment_type = self.get_translated_lines(alignment_trees,
                                                                                                       self.l_from,
                                                                                                       language_to,
                                                                                                       s.get('id'))
                            if translated_lines:
                                translated_sentences = [self.get_line(translation_trees[language_to], line) for line in
                                                        translated_lines]
                                translated_sentences_xml = [self.get_line_as_xml(translation_trees[language_to], line) for line in
                                                            translated_lines]

                                if language_to == 'fr':
                                    for ts in translated_sentences_xml:
                                        for e in ts.xpath(self.config.get(language_to, 'rp_xpath')):
                                            rp = self.check_recent_past(e, language_to)
                                            if rp:
                                                found_trans = True

                                result.append(alignment_type)
                                result.append('<root>' + '\n'.join(translated_sentences) + '</root>' if translated_sentences else '')
                            else:
                                result.append('')
                                result.append('')
                        else:
                            # If no translation is available, add empty columns
                            result.extend([''] * 2)

                    if not found_trans:
                        results.append(result)

        return results

    def get_sentence_words(self, xml_sentence):
        s = []
        # TODO: this xPath-expression might be specific for a corpus
        for w in xml_sentence.xpath('.//w'):
            s.append(w.text.strip() if w.text else ' ')
        return ' '.join(s)


class EuroparlPerfectExtractor(EuroparlExtractor, PerfectExtractor):
    def get_line_and_pp(self, tree, language_to, segment_number):
        """
        Returns the full line for a segment number, as well as the PresentPerfect found (or None if none found).
        TODO: handle more than one here? => bug
        """
        sentence = '-'
        pp = None

        line = self.get_line_as_xml(tree, segment_number)
        if line is not None:
            s = line
            first_w = s.xpath('.//w')[0]
            sentence = get_sentence_from_element(first_w)

            if self.search_in_to:
                for e in s.xpath(self.config.get(language_to, 'xpath')):
                    pp = self.check_present_perfect(e, language_to)
                    if pp:
                        sentence = pp.mark_sentence()
                        break

        return etree.tostring(s, encoding=str), sentence, pp

    def fetch_results(self, filename, s_trees, alignment_trees, translation_trees):
        """
        Processes a single file.
        """
        results = []
        # Find potential present/past perfects (per sentence)
        for _, s in s_trees:
            if self.sentence_ids and s.get('id') not in self.sentence_ids:
                continue

            # Retrieves the xpath expression for the auxiliary in the given tense or a fallback
            xpath_fallback = 'xpath'
            xpath = xpath_fallback + ('_{}'.format(self.tense) if self.tense != PRESENT else '')
            l_config = dict(self.config.items(self.l_from))
            aux_xpath = l_config.get(xpath, l_config.get(xpath_fallback))

            for e in s.xpath(aux_xpath):
                pp = self.check_present_perfect(e, self.l_from)

                # If this is really a present/past perfect, add it to the result
                if pp:
                    result = list()
                    result.append(os.path.basename(filename))
                    result.append(s.get('id'))
                    result.append(pp.perfect_type())
                    result.append(pp.verbs_to_string())
                    result.append(pp.verb_ids())
                    if self.output == XML:
                        result.append('<root>' + etree.tostring(pp.xml_sentence, encoding=str) + '</root>')
                    else:
                        result.append(pp.mark_sentence())
                    self.append_metadata(e, s, result)

                    # Find the translated lines
                    for language_to in self.l_to:
                        if language_to in translation_trees:
                            # TODO: deal with source_lines
                            source_lines, translated_lines, alignment_type = self.get_translated_lines(alignment_trees,
                                                                                                       self.l_from,
                                                                                                       language_to,
                                                                                                       s.get('id'))
                            translated_present_perfects, translated_sentences, translated_marked_sentences = \
                                self.find_translated_present_perfects(translation_trees[language_to], language_to, translated_lines)
                            result.append(alignment_type)
                            if self.output == XML:
                                result.append('<root>' + '\n'.join(translated_sentences) + '</root>' if translated_sentences else '')
                            else:
                                result.append('\n'.join(translated_marked_sentences) if translated_marked_sentences else '')
                        else:
                            # If no translation is available, add empty columns
                            result.extend([''] * 2)

                    results.append(result)

                    # If we want (only) one classification per sentence, break the for loop here.
                    if self.one_per_sentence:
                        break
            else:
                # If we want one classification per sentence, add the sentence with a classification here.
                if self.one_per_sentence:
                    tense, tenses = self.get_tenses(s)

                    result = list()
                    result.append(os.path.basename(filename))
                    result.append(s.get('id'))
                    result.append(tense)
                    result.append(','.join(tenses))
                    result.append('')
                    result.append(self.get_sentence_words(s))
                    self.append_metadata(None, s, result)
                    results.append(result)

        return results



from collections import Counter

from lxml import etree

from perfectextractor.apps.counter.base import BaseCounter
from .base import BaseBNC


class BNCCounter(BaseBNC, BaseCounter):
    def process_file(self, filename):
        """
        Processes a single file.
        """
        results = []

        # Parse the current tree
        tree = etree.parse(filename)
        genre = self.get_genre(tree)

        c = Counter()
        for w in tree.xpath('.//w[@pos="VERB"]'):
            c[w.get('hw', '-')] += 1

        for k, v in c.most_common():
            results.append([filename, genre, k, str(v)])

        return results



# -*- coding: utf-8 -*-

import os
import sys
import unittest

from perfectextractor.corpora.bnc.extractor import BNCPerfectExtractor

DATA_FOLDER = os.path.join(os.path.dirname(__file__), 'data/bnc')
VERBS_COLUMN = 4


class TestBNCPerfectExtractor(unittest.TestCase):
    def setUp(self):
        self.language = 'en'
        self.extractor = BNCPerfectExtractor(self.language)
        self.filename = os.path.join(DATA_FOLDER, 'ALP-formatted.xml')

    def test_init(self):
        self.assertEqual(self.extractor.config.get(self.language, 'perfect_tags'), 'VBN,VDN,VHN,VVN')
        self.assertEqual(self.extractor.config.get(self.language, 'pos'), 'c5')

    def test_list_filenames(self):
        filenames = self.extractor.list_filenames(DATA_FOLDER)
        self.assertEqual(os.path.basename(filenames[0]), 'ALP-formatted.xml')

    def test_process(self):
        results = self.extractor.process_file(self.filename)
        self.assertEqual(len(results), 60)
        self.assertEqual(results[0][VERBS_COLUMN], 'has been presented')
        self.assertEqual(results[1][VERBS_COLUMN], 'has pointed')
        self.assertEqual(results[2][VERBS_COLUMN], 'has shown')
        self.assertEqual(results[3][VERBS_COLUMN], 'has been running')
        self.assertEqual(results[4][VERBS_COLUMN], 'has devoted')

    def test_ppc(self):
        # Test whether present perfect continuous are ignored when check_ppc is set to False
        # Only works on Python 3 for some reason...
        if sys.version_info > (3,):
            self.extractor.config[self.language]['ppc'] = 'false'
            results = self.extractor.process_file(self.filename)
            self.assertEqual(results[3][VERBS_COLUMN], 'has been')

    def test_lemmata_list(self):
        # Test whether the lemmata_list will exclude the listed verbs
        self.extractor = BNCPerfectExtractor(self.language, lemmata=['show', 'run'])
        results = self.extractor.process_file(self.filename)
        self.assertEqual(results[0][VERBS_COLUMN], 'has shown')
        self.assertEqual(results[1][VERBS_COLUMN], 'has been running')
        self.assertEqual(results[2][VERBS_COLUMN], 'has shown')
        self.assertEqual(results[3][VERBS_COLUMN], 'have been shown')
        self.assertEqual(results[4][VERBS_COLUMN], 'has shown')

    def test_not_implemented(self):
        self.assertRaises(NotImplementedError, self.extractor.get_line_and_pp, None, None, None)
        self.assertRaises(NotImplementedError, self.extractor.get_translated_lines, None, None, None, None)



from abc import ABCMeta, abstractmethod
import codecs
import csv
import os

import click

from .utils import TXT

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser

LEMMATA_CONFIG = os.path.join(os.path.dirname(__file__), 'config/{language}_lemmata.txt')


class CachedConfig:
    def __init__(self, config):
        self.cache = dict()
        self.config = config

    def setdefault(self, key, func):
        if key in self.cache:
            return self.cache[key]
        return self.cache.setdefault(key, func())

    def get(self, section, key):
        return self.setdefault((section, key), lambda: self.config.get(section, key))

    def getboolean(self, section, key):
        return self.setdefault((section, key), lambda: self.config.getboolean(section, key))

    def items(self, section):
        return self.setdefault(section, lambda: self.config.items(section))

    def __getitem__(self, key):
        return self.config[key]


class BaseExtractor(object):
    __metaclass__ = ABCMeta

    def __init__(self, language_from, languages_to=None,
                 file_names=None, sentence_ids=None,
                 lemmata=None, tokens=None, metadata=None, regex=None,
                 outfile=None, position=None, output=TXT,
                 one_per_sentence=False,
                 sort_by_certainty=False, file_limit=0, min_file_size=0, max_file_size=0):
        """
        Initializes the extractor for the given source and target language(s).
        :param language_from: the source language
        :param languages_to: the target language(s)
        :param file_names: whether to limit the search to certain file names
        :param sentence_ids: whether to limit the search to certain sentence IDs
        :param lemmata: whether to limit the search to certain lemmata (can be provided as a boolean or a list)
        :param tokens: whether to limit the search to certain tokens (list of tuples (from-to))
        :param metadata: whether to add metadata to the output (list of tuples (metadata-level))
        :param regex: whether to limit the search to certain regular expressions (list of regexes)
        :param outfile: the filename to output the results to
        :param position: whether to limit the search to a certain position (e.g. only sentence-initial)
        :param output: whether to output the results in text or XML format
        :param one_per_sentence: whether to output all lines, and allow one classification per sentence
        :param sort_by_certainty: whether to sort the files by average alignment certainty
        :param file_limit: whether to limit the number of files searched in
        :param min_file_size: whether to only use files larger (or equal) than a certain size
        :param max_file_size: whether to only use files smaller (or equal) than a certain size
        """
        self.l_from = language_from
        self.l_to = languages_to or []
        self.file_names = file_names
        self.sentence_ids = sentence_ids
        self.tokens = dict(tokens) if tokens else None
        self.metadata = dict(metadata) if metadata else {}
        self.regex = regex
        self.outfile = outfile
        self.position = position
        self.output = output
        self.one_per_sentence = one_per_sentence
        self.sort_by_certainty = sort_by_certainty
        self.file_limit = file_limit
        self.min_file_size = min_file_size
        self.max_file_size = max_file_size

        # Read in the lemmata list (if provided)
        self.lemmata_list = []
        self.read_lemmata(lemmata)

        # Read the config)
        with codecs.open(self.get_config(), 'r', 'utf8') as config_file:
            config = ConfigParser.ConfigParser()
            try:
                config.read_file(config_file)
            except AttributeError:  # Support for old method (Python < 3.2)
                config.readfp(config_file)
            self.config = CachedConfig(config)

        # Other variables
        self.other_extractors = []
        self.alignment_xmls = dict()

    def process_folder(self, dir_name, progress_cb=None, done_cb=None):
        """
        Creates a result file and processes each file in a folder.
        """
        file_names = self.collect_file_names(dir_name)
        progress_total = len(file_names)

        result_file = self.outfile or '-'.join([dir_name, self.l_from]) + '.csv'
        with open(result_file, 'w') as f:
            f.write('\uFEFF')  # the UTF-8 BOM to hint Excel we are using that...
            csv_writer = csv.writer(f, delimiter=';')
            csv_writer.writerow(self.generate_header())
            for i, part in enumerate(self.generate_results(dir_name, file_names)):
                csv_writer.writerows(part)
                if progress_cb:
                    progress_cb(i + 1, progress_total)

        if done_cb:
            done_cb(result_file)

    def collect_file_names(self, dir_name):
        """
        Collects the file names in a given directory and (potentially) filters these based on file size,
        alignment certainty or a limited number of files.
        :param dir_name: The current directory
        :return: A list of files to consider.
        """
        click.echo('Collecting file names...')

        if self.file_names:
            file_names = [os.path.join(dir_name, f) for f in self.file_names]
        else:
            file_names = self.list_filenames(dir_name)

        if self.min_file_size or self.max_file_size:
            file_names = self.filter_by_file_size(file_names)

        if self.sort_by_certainty:
            file_names = self.sort_by_alignment_certainty(file_names)

        if self.file_limit:
            file_names = file_names[:self.file_limit]

        click.echo('Finished collecting file names, starting processing...')
        return file_names

    def generate_results(self, dir_name, file_names=None):
        if file_names is None:
            file_names = self.collect_file_names(dir_name)

        for f in file_names:
            yield self.process_file(f)

    def generate_header(self):
        header = [
            'document',
            'sentence',
            'type {}'.format(self.l_from),
            'words {}'.format(self.l_from),
            'ids {}'.format(self.l_from),
            self.l_from]
        for metadata in self.metadata.keys():
            header.append(metadata)
        for language in self.l_to:
            header.append('alignment type')
            header.append(language)
        return header

    def read_lemmata(self, lemmata):
        if lemmata is not None:
            if type(lemmata) in (list, tuple):
                self.lemmata_list = list(lemmata)
            elif type(lemmata) == bool:
                if lemmata:
                    with codecs.open(LEMMATA_CONFIG.format(language=self.l_from), 'r', 'utf-8') as lexicon:
                        self.lemmata_list = lexicon.read().split()
            else:
                raise ValueError('Unknown value for lemmata')

    def add_extractor(self, extractor):
        """
        Adds another Extractor to this Extractor. This allows to combine Extractors.
        The last added Extractor determines the output.
        """
        self.other_extractors.append(extractor)

    def list_directories(self, path):
        directories = [os.path.join(path, directory) for directory in os.listdir(path)]
        return filter(os.path.isdir, directories)

    def append_metadata(self, w, s, result):
        for metadata, level in self.metadata.items():
            if w is not None and level == 'w':
                result.append(w.get(metadata))
            elif s is not None and level == 's':
                result.append(s.get(metadata))
            elif s is not None and level == 'p':
                result.append(s.getparent().get(metadata))
            elif s is not None and level == 'text':
                result.append(s.getparent().getparent().get(metadata))
            else:
                raise ValueError('Invalid level {}'.format(level))

    def get_pos(self, language, element):
        return element.get(self.config.get(language, 'pos'))

    def get_tenses(self, sentence):
        """
        This method allows to retrieve the "tense" for a sentence. It is very naive,
        based upon the part-of-speech tags of verbs that appear in the sentence.
        It should work for the tagsets of both the Penn Treebank Project and the BNC.
        See https://www.cis.uni-muenchen.de/~schmid/tools/TreeTagger/data/Penn-Treebank-Tagset.pdf for the
        Penn Treebank Project tagset and see http://www.natcorp.ox.ac.uk/docs/URG/posguide.html#section1
        for the BNC tagset
        :param sentence: the s element
        :return: a tuple of the assigned tense and all tenses for the verbs in the sentences
        """
        tense = 'none'
        tenses = []
        for w in sentence.xpath('.//w'):
            pos = self.get_pos(self.l_from, w)

            if pos.startswith('V') and len(pos) == 3:
                if pos.endswith('B') or pos.endswith('P') or pos.endswith('Z'):
                    tenses.append('present')
                elif pos.endswith('D'):
                    tenses.append('past')
                elif pos.endswith('N'):
                    tenses.append('participle')
                elif pos.endswith('G'):
                    tenses.append('gerund')
                elif pos.endswith('I'):
                    tenses.append('infinitive')
                elif pos == 'VM0':
                    tenses.append('modal')
            elif pos == 'MD':
                tenses.append('modal')
            elif pos == 'BES':
                tenses.append('present')
            elif pos == 'VB':
                tenses.append('infinitive')

        if tenses:
            tenses_set = set(tenses)
            if len(tenses_set) == 1:
                tense = tenses[0]
            else:
                if tenses_set in [{'present', 'infinitive'}, {'present', 'gerund'}, {'present', 'gerund', 'infinitive'}]:
                    tense = 'present'
                elif tenses_set in [{'past', 'infinitive'}, {'past', 'gerund'}, {'past', 'gerund', 'infinitive'}]:
                    tense = 'past'
                elif tenses_set == {'modal', 'infinitive'}:
                    tense = 'modal'
                else:
                    tense = 'other'

        return tense, tenses

    @abstractmethod
    def get_config(self):
        """
        Returns the location of the configuration file.
        """
        raise NotImplementedError

    @abstractmethod
    def process_file(self, filename):
        """
        Process a single file
        """
        raise NotImplementedError

    @abstractmethod
    def list_filenames(self, dir_name):
        """
        List all to be processed files in the given directory.
        """
        raise NotImplementedError

    @abstractmethod
    def get_translated_lines(self, alignment_trees, language_from, language_to, segment_number):
        """
        Returns the translated segment numbers (could be multiple) for a segment number in the original text.
        """
        raise NotImplementedError

    @abstractmethod
    def get_sentence(self, element):
        """
        Returns the full sentence XML for the given element.
        """
        raise NotImplementedError

    @abstractmethod
    def get_siblings(self, element, sentence_id, check_preceding):
        """
        Returns the siblings of the given element in the given sentence_id.
        The check_preceding parameter allows to look either forwards or backwards.
        """
        raise NotImplementedError

    @abstractmethod
    def sort_by_alignment_certainty(self, file_names):
        raise NotImplementedError

    @abstractmethod
    def filter_by_file_size(self, file_names):
        raise NotImplementedError



import glob
import os

EUROPARL_CONFIG = os.path.join(os.path.dirname(__file__), 'europarl.cfg')


class BaseEuroparl(object):
    def get_config(self):
        return EUROPARL_CONFIG

    def list_filenames(self, dir_name):
        return sorted(glob.glob(os.path.join(dir_name, '*.xml')))
